/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>

#include <pan/base/base-prefs.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-config.h>
#include <pan/base/serverlist.h>

#include <pan/pan-file-entry.h>
#include <pan/server-ui.h>
#include <pan/server-menu-ui.h>
#include <pan/util.h>

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

typedef struct
{
	Server * server;
	gboolean new_server;

	GtkWidget * dialog;

	GtkWidget * name_entry;
	GtkWidget * address_entry;
	GtkWidget * port_spin;
	GtkWidget * connection_limit_spin;
	GtkWidget * reserve_reading_socket_checkbox;

	GtkWidget * auth_checkbox;
	GtkWidget * auth_username_label;
	GtkWidget * auth_username_entry;
	GtkWidget * auth_password_label;
	GtkWidget * auth_password_entry;

	GtkWidget * newsrc_checkbox;
	GtkWidget * newsrc_filename_label;
	GtkWidget * newsrc_filename_entry;
}
ServerEditDialog;

typedef struct
{
	GtkWidget * server_clist;
	GtkWidget * dialog;

	GtkWidget * remove_button;
	GtkWidget * edit_button;
}
ServerListDialog;

/*********************
**********************  Private Function Prototypes
*********************/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  EDIT DIALOG
************/

static void
edit_dialog_sensitize_newsrc (ServerEditDialog * d, gboolean enabled)
{
	gtk_widget_set_sensitive (d->newsrc_filename_label, enabled);
	gtk_widget_set_sensitive (d->newsrc_filename_entry, enabled);
}

static void
edit_dialog_newsrc_toggled_cb (GtkToggleButton * tb, gpointer user_data)
{
	ServerEditDialog * d = (ServerEditDialog*) user_data;
	gboolean checked = gtk_toggle_button_get_active (tb);
	edit_dialog_sensitize_newsrc (d, checked);
}

static void
edit_dialog_sensitize_authorization (ServerEditDialog * d, gboolean enabled)
{
	gtk_widget_set_sensitive (d->auth_username_label, enabled);
	gtk_widget_set_sensitive (d->auth_username_entry, enabled);
	gtk_widget_set_sensitive (d->auth_password_label, enabled);
	gtk_widget_set_sensitive (d->auth_password_entry, enabled);
}

static void
edit_dialog_auth_toggled_cb (GtkToggleButton * tb, gpointer user_data)
{
	ServerEditDialog * d = (ServerEditDialog*) user_data;
	gboolean checked = gtk_toggle_button_get_active (tb);
	edit_dialog_sensitize_authorization (d, checked);
}

static void
edit_dialog_populate (ServerEditDialog * d, Server * s)
{
	GtkAdjustment * a;
	const char * cpch;

	/* sanity clause */
	g_return_if_fail (d!=NULL);
	g_return_if_fail (GTK_IS_WIDGET(d->dialog));
	g_return_if_fail (server_is_valid(s));

	/* internal bits */
	d->server = s;

	/* server section */
	cpch = s->name;
	gtk_entry_set_text (GTK_ENTRY(d->name_entry), cpch ? cpch : "");
	cpch = s->address;
	gtk_entry_set_text (GTK_ENTRY(d->address_entry), cpch ? cpch : "");
	a = gtk_spin_button_get_adjustment (GTK_SPIN_BUTTON(d->port_spin));
	gtk_adjustment_set_value (a, s->port);
	a = gtk_spin_button_get_adjustment (GTK_SPIN_BUTTON(d->connection_limit_spin));
	gtk_adjustment_set_value (a, s->max_connections);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(d->reserve_reading_socket_checkbox), s->reserve_connection_for_bodies);

	/* authorization */
	cpch = s->username;
	gtk_entry_set_text (GTK_ENTRY(d->auth_username_entry), cpch ? cpch : "");
	cpch = s->password;
	gtk_entry_set_text (GTK_ENTRY(d->auth_password_entry), cpch ? cpch : "");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(d->auth_checkbox), s->need_auth);
	edit_dialog_sensitize_authorization (d, s->need_auth);

	/* newsrc */
	pan_file_entry_set (d->newsrc_filename_entry, s->newsrc_filename);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(d->newsrc_checkbox), s->use_newsrc);
	edit_dialog_sensitize_newsrc (d, s->use_newsrc);
}

static void
edit_dialog_populate_server (ServerEditDialog * d, Server * s)
{
	gboolean b;
	const char * cpch;

	/* sanity clause */
	g_return_if_fail (d!=NULL);
	g_return_if_fail (GTK_IS_WIDGET(d->dialog));
	g_return_if_fail (server_is_valid(s));

	/* server section */
	cpch = gtk_entry_get_text (GTK_ENTRY(d->name_entry));
	replace_gstr (&s->name, g_strdup(cpch));
	g_strstrip (s->name);
	cpch = gtk_entry_get_text (GTK_ENTRY(d->address_entry));
	replace_gstr (&s->address, g_strdup(cpch));
	g_strstrip (s->address);
	s->port = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(d->port_spin));
	s->max_connections = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(d->connection_limit_spin));
	b = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(d->reserve_reading_socket_checkbox));
	s->reserve_connection_for_bodies = b;

	/* authorization */
	b = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(d->auth_checkbox));
	s->need_auth = b;
	cpch = gtk_entry_get_text (GTK_ENTRY(d->auth_username_entry));
	replace_gstr (&s->username, g_strdup(cpch));
	g_strstrip (s->username);
	cpch = gtk_entry_get_text (GTK_ENTRY(d->auth_password_entry));
	replace_gstr (&s->password, g_strdup(cpch));

	/* newsrc */
	b = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(d->newsrc_checkbox));
	s->use_newsrc = b;
	cpch = pan_file_entry_get (d->newsrc_filename_entry);
	replace_gstr (&s->newsrc_filename, g_strdup(cpch));
	g_strstrip (s->newsrc_filename);
}

static void
server_edit_response_cb (GtkDialog * w, int response, gpointer user_data)
{
	ServerEditDialog * d = (ServerEditDialog*) user_data;

	if (response == GTK_RESPONSE_OK)
	{
		edit_dialog_populate_server (d, d->server);

		if (d->new_server)
			serverlist_add_server (d->server);
	}
	else /* cancel, destroy */
	{
		if (d->new_server)
			pan_object_unref (PAN_OBJECT(d->server));
	}

	gtk_widget_destroy (GTK_WIDGET(w));
}

static GtkWidget*
server_edit_dialog (GtkWidget * window, Server * server)
{
	char buf[512];
	GtkObject * a;
	GtkWidget * dialog;
	GtkWidget * w;
	GtkWidget * t;
	GtkWidget * l;
	ServerEditDialog * d;
	int row;
	const int max_per_server = pan_config_get_int ("/Pan/General/Max_Connections_Per_Server=4");

	d = g_new0 (ServerEditDialog, 1);

	/* create the dialog */
	dialog = gtk_dialog_new_with_buttons (_("Pan: Edit Server"),
	                                      GTK_WINDOW(window),
	                                      GTK_DIALOG_DESTROY_WITH_PARENT,
	                                      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	                                      GTK_STOCK_OK, GTK_RESPONSE_OK,
	                                      NULL);
	d->dialog = dialog;
	g_object_set_data_full (G_OBJECT(dialog), "dialog", d, g_free);
	g_signal_connect (dialog, "response", G_CALLBACK(server_edit_response_cb), d);

	/**
	***  workarea
	**/

	row = 0;
	t = gtk_table_new (4, 18, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(t), 12);
	gtk_table_set_row_spacings (GTK_TABLE(t), 6);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(d->dialog)->vbox), t, TRUE, TRUE, 0);

		/* title for name section */
		g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("Server"));
		l = gtk_label_new (buf);
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_label_set_use_markup (GTK_LABEL(l), TRUE);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 0, 4, row, row+1);
		++row;

		/* spacer to move the fields a little to the right of the name header */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 18u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 1, row, row+5);

		/* spacer between the controls and their labels */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 12u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 2, 3, row, row+5);

		l = gtk_label_new_with_mnemonic (_("_Name:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		w = d->name_entry = gtk_entry_new ();
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

		l = gtk_label_new_with_mnemonic (_("_Address:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		w = d->address_entry = gtk_entry_new ();
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

		l = gtk_label_new_with_mnemonic (_("Por_t:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		a = gtk_adjustment_new (1.0, 1.0, ULONG_MAX, 1.0, 1.0, 1.0);
		w = d->port_spin = gtk_spin_button_new (GTK_ADJUSTMENT(a), 1.0, 0u);
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

	/* spacer between "name" and "connections" */
	w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
	gtk_widget_set_usize (w, 0u, 6u); 
	gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 4, row, row+1);
	++row;

		/* title for name section */
		g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("Connections"));
		l = gtk_label_new (buf);
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_label_set_use_markup (GTK_LABEL(l), TRUE);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 0, 4, row, row+1);
		++row;

		/* spacer to move the fields a little to the right of the section header */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 18u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 1, row, row+3);

		/* spacer between the controls and their labels */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 12u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 2, 3, row, row+3);

		w = gtk_check_button_new_with_mnemonic (_("_Reserve a connection for reading articles"));
		d->reserve_reading_socket_checkbox = w;
		gtk_table_attach_defaults (GTK_TABLE(t), w, 1, 4, row, row+1);
		++row;

		l = gtk_label_new_with_mnemonic (_("Connection _Limit:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		a = gtk_adjustment_new (1.0, 1.0, max_per_server, 1.0, 1.0, 1.0);
		w = gtk_spin_button_new (GTK_ADJUSTMENT(a), 1.0, 0u);
		d->connection_limit_spin = w;
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

	/* spacer between "connections" and "authorization" */
	w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
	gtk_widget_set_usize (w, 0u, 6u); 
	gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 4, row, row+1);
	++row;

		/* title for name section */
		g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("Authorization"));
		l = gtk_label_new (buf);
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_label_set_use_markup (GTK_LABEL(l), TRUE);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 0, 4, row, row+1);
		++row;

		/* spacer to move the fields a little to the right of the section header */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 18u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 1, row, row+3);

		/* spacer between the controls and their labels */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 12u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 2, 3, row, row+3);

		w = d->auth_checkbox = gtk_check_button_new_with_mnemonic (_("Server requires authori_zation"));
		gtk_table_attach_defaults (GTK_TABLE(t), w, 1, 4, row, row+1);
		g_signal_connect (w, "toggled", G_CALLBACK(edit_dialog_auth_toggled_cb), d);
		++row;

		l = d->auth_username_label = gtk_label_new_with_mnemonic (_("_Username:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		w = d->auth_username_entry = gtk_entry_new ();
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

		l = d->auth_password_label = gtk_label_new_with_mnemonic (_("_Password:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		w = d->auth_password_entry = gtk_entry_new ();
		gtk_entry_set_visibility (GTK_ENTRY(w), FALSE);
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

	/* spacer between "authorization" and "newsrc" */
	w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
	gtk_widget_set_usize (w, 0u, 6u); 
	gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 4, row, row+1);
	++row;

		/* title for name section */
		g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("Newsrc"));
		l = gtk_label_new (buf);
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_label_set_use_markup (GTK_LABEL(l), TRUE);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 0, 4, row, row+1);
		++row;

		/* spacer to move the fields a little to the right of the section header */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 18u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 0, 1, row, row+2);

		/* spacer between the controls and their labels */
		w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
		gtk_widget_set_usize (w, 12u, 0u); 
		gtk_table_attach_defaults (GTK_TABLE(t), w, 2, 3, row, row+2);

		w = d->newsrc_checkbox = gtk_check_button_new_with_mnemonic (_("Us_e .newsrc file"));
		gtk_table_attach_defaults (GTK_TABLE(t), w, 1, 4, row, row+1);
		g_signal_connect (w, "toggled", G_CALLBACK(edit_dialog_newsrc_toggled_cb), d);
		++row;

		l = d->newsrc_filename_label = gtk_label_new_with_mnemonic (_("_Filename:"));
		gtk_misc_set_alignment (GTK_MISC(l), 0.0f, 0.5f);
		gtk_table_attach_defaults (GTK_TABLE(t), l, 1, 2, row, row+1);

		w = d->newsrc_filename_entry = pan_file_entry_new (_("Select .newsrc File"));
		gtk_widget_set_usize (w, 300u, 0u); 
		gtk_label_set_mnemonic_widget (GTK_LABEL(l), pan_file_entry_gtk_entry(w));
		gtk_table_attach_defaults (GTK_TABLE(t), w, 3, 4, row, row+1);
		++row;

	if (server == NULL) {
		d->new_server = TRUE;
		server = server_new ();
		server->name = g_strdup (_("New Server"));
		server->port = 119;
	}
	edit_dialog_populate (d, server);
	gtk_widget_show_all (d->dialog);
	return d->dialog;
}


/************
*************  LIST DIALOG
************/

static int
get_selected_index (ServerListDialog * d)
{
	int sel;
	GtkCList * clist;

	/* find the selected server */
	pan_lock ();
	clist = GTK_CLIST(d->server_clist);
	sel = -1;
	if (clist->selection != NULL)
		sel = GPOINTER_TO_INT(clist->selection->data);
	pan_unlock ();

	return sel;
}

static Server*
get_selected_server (ServerListDialog * d)
{
	Server * retval = NULL;

	const int index = get_selected_index (d);
	if (index != -1)
		retval = (Server*) gtk_clist_get_row_data (GTK_CLIST(d->server_clist), index);

	return retval;
}

static void
button_refresh (ServerListDialog * d)
{
	const int sel = get_selected_index (d);
	const gboolean have_sel = sel != -1;

	pan_lock ();
	gtk_widget_set_sensitive (d->edit_button, have_sel);
	gtk_widget_set_sensitive (d->remove_button, have_sel);
	pan_unlock ();
}

static int
compare_ppserver_ppserver_by_name (gconstpointer a, gconstpointer b)
{
	const Server * s1 = *(const Server **)a;
	const Server * s2 = *(const Server **)b;
	return pan_strcmp (s1->name, s2->name);
}

static GPtrArray*
get_public_server_array (void)
{
	int i;
	GPtrArray * a = g_ptr_array_new ();
	serverlist_get_servers (a);

	/* skip the internal folder... */
	for (i=0; i<a->len; ++i)
		if (!pan_strcmp (INTERNAL_SERVER_NAME, SERVER(g_ptr_array_index(a,i))->name))
			break;
	if (i!=a->len)
		g_ptr_array_remove_index (a, i);
	g_ptr_array_sort (a, compare_ppserver_ppserver_by_name);

	return a;
}

static void
clist_refresh (ServerListDialog * d)
{
	int i;
	int select_row = -1;
	Server * selected_server = get_selected_server (d);
	GtkCList * clist = GTK_CLIST(d->server_clist);
	GPtrArray * a = get_public_server_array ();

	pan_lock ();
	gtk_clist_freeze (clist);
	gtk_clist_clear (clist);
	for (i=0; i<a->len; ++i)
	{
		int row;
		Server * server = SERVER (g_ptr_array_index (a, i));
		row = gtk_clist_insert (clist, -1, &server->name);
		gtk_clist_set_row_data (clist, row, server);
		if (server == selected_server)
			select_row = i;
	}
	if (select_row != -1)
		gtk_clist_select_row (clist, select_row, 0);
	gtk_clist_thaw (clist);
	pan_unlock ();

	g_ptr_array_free (a, TRUE);
}

/*****
******
*****/

static void
server_list_dialog_response_cb (GtkDialog * dialog, int response, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(dialog));
}

static void
server_list_dialog_destroy_cb (GtkWidget * w, gpointer user_data)
{
	g_free (user_data);

	server_menu_update ();
}


/**
***  Add
**/

static void
remove_button_clicked_cb (GtkButton * button, gpointer data)
{
	ServerListDialog * d = (ServerListDialog*) data;
	Server * selected_server = get_selected_server (d);
	if (selected_server != NULL)
	{
		serverlist_remove_server (selected_server);
		clist_refresh (d);
		button_refresh (d);
	}
}

static void
server_edit_dialog_destroy_cb (GtkWidget * w, gpointer user_data)
{
	ServerListDialog * d = (ServerListDialog*) user_data;
	clist_refresh (d);
}

static void
add_button_clicked_cb (GtkButton * button, gpointer data)
{
	ServerListDialog * f = (ServerListDialog*) data;
	GtkWidget * dialog = server_edit_dialog (f->dialog, NULL);
	g_signal_connect (dialog, "destroy", G_CALLBACK(server_edit_dialog_destroy_cb), f);
	gtk_widget_show_all (dialog);
}

/**
***  Edit
**/

static void
edit_button_clicked_cb (GtkButton * button, gpointer user_data)
{
	ServerListDialog * d = (ServerListDialog*) user_data;
	Server * selected_server = get_selected_server (d);
	if (selected_server != NULL)
	{
		GtkWidget * dialog = server_edit_dialog (d->dialog, selected_server);
		g_signal_connect (GTK_OBJECT(dialog), "destroy", G_CALLBACK(server_edit_dialog_destroy_cb), d);
		gtk_widget_show_all (dialog);
	}
}
static gboolean
server_clist_button_press_cb (GtkWidget * w, GdkEventButton * b, gpointer data)
{
	if (b->button==1 && b->type==GDK_2BUTTON_PRESS)
		edit_button_clicked_cb (NULL, data);
	return FALSE;
}

static void     
list_selection_changed_cb (GtkCList          * clist,
                           int                 row,
                           int                 column,
                           GdkEventButton    * event,
                           gpointer            user_data)
{
	button_refresh ((ServerListDialog*)user_data);
}

/************
*************  PUBLIC
************/

GtkWidget*
server_dialog_new (GtkWindow * parent)
{
	GtkWidget * w;
	GtkWidget * hbox;
	GtkWidget * bbox;
	GtkTooltips * tips = gtk_tooltips_new ();
	ServerListDialog * d = g_new0 (ServerListDialog, 1);
	char * titles [1];

	/* dialog */
	w = d->dialog = gtk_dialog_new_with_buttons (_("Pan: Servers"),
	                                             parent,
	                                             GTK_DIALOG_DESTROY_WITH_PARENT,
	                                             GTK_STOCK_CLOSE, GTK_RESPONSE_OK,
	                                             NULL);
	gtk_window_set_policy (GTK_WINDOW(w), TRUE, TRUE, TRUE);
	g_signal_connect (GTK_OBJECT(w), "response", G_CALLBACK(server_list_dialog_response_cb), d);
	g_signal_connect (GTK_OBJECT(w), "destroy", G_CALLBACK(server_list_dialog_destroy_cb), d);

	/* workarea */
	hbox = gtk_hbox_new (FALSE, GUI_PAD);
	gtk_container_set_border_width (GTK_CONTAINER(hbox), 12);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(w)->vbox), hbox, TRUE, TRUE, 0);

	/* clist */
	titles[0] = _("Servers");
	w = d->server_clist= gtk_clist_new_with_titles (1, titles);
	gtk_clist_set_selection_mode (GTK_CLIST(w), GTK_SELECTION_BROWSE);
	g_signal_connect (w, "button_press_event",
	                  G_CALLBACK(server_clist_button_press_cb), d);
	g_signal_connect (w, "select_row",
	                  G_CALLBACK(list_selection_changed_cb), d);
	g_signal_connect (w, "unselect_row",
	                  G_CALLBACK(list_selection_changed_cb), d);


        w = gtk_scrolled_window_new (NULL, NULL);
	gtk_container_add (GTK_CONTAINER(w), d->server_clist);
	gtk_box_pack_start (GTK_BOX (hbox), w, TRUE, TRUE, 0);
	gtk_widget_set_usize (w, 300, 300);

	/* button box */
	bbox = gtk_vbox_new (FALSE, GUI_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (hbox), bbox, FALSE, FALSE, 0);

	/* add button */
	w = gtk_button_new_from_stock (GTK_STOCK_ADD);
	gtk_box_pack_start (GTK_BOX (bbox), w, FALSE, FALSE, 0);
	gtk_tooltips_set_tip (tips, w, _("Add a new server"), NULL);
	g_signal_connect (w, "clicked", G_CALLBACK(add_button_clicked_cb), d);

	/* edit button */
	w = gtk_button_new_from_stock (GTK_STOCK_OPEN);
	gtk_box_pack_start (GTK_BOX (bbox), w, FALSE, FALSE, 0);
	gtk_tooltips_set_tip (tips, w, _("Edit the selected server"), NULL);
	g_signal_connect (w, "clicked", G_CALLBACK(edit_button_clicked_cb), d);
	d->edit_button = w;

	/* remove button */
	w = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
	gtk_box_pack_start (GTK_BOX (bbox), w, FALSE, FALSE, 0);
	gtk_tooltips_set_tip (tips, w, _("Remove the selected server"), NULL);
	g_signal_connect (w, "clicked", G_CALLBACK(remove_button_clicked_cb), d);
	d->remove_button = w;

	clist_refresh (d);
	button_refresh (d);
	return d->dialog;
}
